import json
import subprocess
import unittest

import boto3
import yaml


class TestAWSModules(unittest.TestCase):
    def setUp(self):
        s3 = subprocess.Popen(
            ['moto_server', 's3', '-p', '5000'],
            stderr=subprocess.PIPE)
        sqs = subprocess.Popen(
            ['moto_server', 'sqs', '-p', '5001'],
            stderr=subprocess.PIPE)
        s3.stderr.readline()
        sqs.stderr.readline()
        self.s3 = s3
        self.sqs = sqs
        with open('tests/aws/env.yml') as f:
            y = yaml.load(f.read())
        cbc = {'LocationConstraint': y['s3_config']['region_name']}
        boto_s3 = boto3.resource('s3', **y['s3_config'])
        boto_sqs = boto3.resource('sqs', **y['sqs_config'])
        boto_s3.create_bucket(Bucket='sample-bucket',
                              CreateBucketConfiguration=cbc)
        boto_sqs.create_queue(QueueName='sample-queue')

    def tearDown(self):
        s3 = self.s3
        sqs = self.sqs
        s3.kill()
        sqs.kill()
        s3.stderr.close()
        sqs.stderr.close()
        s3.wait()
        sqs.wait()

    def run_recipe(self, taskname):
        c = subprocess.run([
            'python3',
            '-m', 'bamboo_crawler',
            '-r', 'tests/aws/recipe.yml',
            '-e', 'tests/aws/env.yml',
            '-t', taskname],
            stdout=subprocess.PIPE)
        self.assertEqual(c.returncode, 0)
        return c

    def test_aws_inputter_and_outputter(self):
        self.run_recipe('aws_outputter')
        c = self.run_recipe('aws_inputter')
        j = json.loads(c.stdout)
        self.assertEqual(j['sampledata'], 'ABCDEFGHIJKLMN')



import json
import pathlib
import sys
import time
from types import MappingProxyType

import boto3
import sqlalchemy
from sqlalchemy.ext.automap import automap_base

from .. import interface


class ConstantInputter(interface.Inputter):
    def __init__(self, value, metadata=MappingProxyType({})):
        self.value = value
        self.metadata = metadata

    def read(self):
        return interface.Context(self.value, **self.metadata)


class StdinInputter(interface.Inputter):
    def read(self):
        return interface.Context(sys.stdin.read())


class FileInputter(interface.Inputter):
    def __init__(self, filepath):
        self.path = pathlib.Path(filepath).resolve()

    def read(self):
        return interface.Context(self.path.open().read())


class SQSS3Inputter(interface.Inputter):

    def __init__(self, bucket_name, queue_name,
                 *,
                 s3_config=MappingProxyType({}),
                 sqs_config=MappingProxyType({})):
        s3 = boto3.resource('s3', **s3_config)
        sqs = boto3.resource('sqs', **sqs_config)
        bucket = s3.Bucket(bucket_name)
        queue = sqs.get_queue_by_name(QueueName=queue_name)
        self.bucket = bucket
        self.queue = queue
        self.message = None

    def read(self):
        message = self.__read_sqs()
        body, meta = self.__read_s3(message)
        return interface.Context(body, **meta)

    def __read_sqs(self):
        if self.message is not None:
            return self.message

        messages = list(self.queue.receive_messages())
        while not messages:
            time.sleep(5)
            messages = list(self.queue.receive_messages())

        message = messages[0]
        self.message = message
        return message

    def __read_s3(self, message):
        j = json.loads(message.body)
        obj = self.bucket.Object(j['s3_body'])
        metadata = j['metadata']
        response = obj.get()
        return response['Body'].read(), metadata

    def done(self):
        super().done()
        self.message.delete()
        self.message = None


class SQLInputter(interface.Inputter):

    def __init__(self, url, *, table=None, query=None):
        if query is None and table is None:
            raise NotImplementedError
        if query is not None and table is not None:
            raise NotImplementedError
        self.engine = sqlalchemy.create_engine(url)
        if table is not None:
            metadata = sqlalchemy.MetaData()
            metadata.reflect(self.engine, only=[table])
            Base = automap_base(metadata=metadata)
            Base.prepare()
            query = metadata.tables[table].select()
        self.result = self.engine.execute(query)
        self.keys = self.result.keys()

    def read(self):
        r = self.result.fetchone()
        p = dict(zip(self.keys, r))
        return interface.Context(p)


__all__ = [
    'ConstantInputter',
    'StdinInputter',
    'FileInputter',
    'SQSS3Inputter',
    'SQLInputter',
]



import argparse
import logging
import sys
import time

import yaml

from . import deserializers
from . import inputters
from . import outputters
from .parser import parse_recipe
from . import processors
from . import serializers
from .task import Task


def _set_logger(debug=False):
    logger = logging.getLogger(__name__)
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.ERROR)
    fmt = logging.Formatter(logging.BASIC_FORMAT)
    ch = logging.StreamHandler(sys.stderr)
    ch.setFormatter(fmt)
    logger.addHandler(ch)
    return logger


def pickup_class(directive, search_classes):
    return getattr(search_classes, directive['type'])


def subtask_from_directive(directive, search_classes):
    options = directive.get('options', {})
    class_ = pickup_class(directive, search_classes)
    target = class_(**options)
    return target


def task_from_definition(name, inputters, outputters, processors, definitions):
    d = definitions
    reserved_keywords = frozenset({'input', 'output',
                                   'process',
                                   'deserialize', 'serialize'})
    if frozenset(d.keys()) > reserved_keywords:
        raise NotImplementedError

    inputter = subtask_from_directive(d['input'], inputters)
    outputter = subtask_from_directive(d['output'], outputters)
    processor = subtask_from_directive(d['process'], processors)

    if 'deserialize' in d:
        deserializer = subtask_from_directive(d['deserialize'], deserializers)
    else:
        deserializer = None

    if 'serialize' in d:
        serializer = subtask_from_directive(d['serialize'], serializers)
    else:
        serializer = None

    task = Task(name=name,
                inputter=inputter,
                processor=processor,
                outputter=outputter,
                deserializer=deserializer,
                serializer=serializer)
    return task


def job_(recipe, tasks, duration):
    logger = logging.getLogger(__name__)
    for key in tasks:
        definitions = recipe[key]
        task = task_from_definition(key, inputters, outputters,
                                    processors, definitions)
        logger.debug('begin: ' + key)
        task.do()
        logger.debug('end: ' + key)
        time.sleep(duration)


def job(envs, infile, tasks, show_only, duration, loop=False):
    recipe = parse_recipe(infile, envs)
    if not tasks:
        tasks = recipe.keys()
    if show_only:
        for key in tasks:
            print(key)
        return
    for t in tasks:
        if t not in recipe.keys():
            raise NotImplementedError
    if loop:
        while True:
            job_(recipe, tasks, duration)
    else:
        job_(recipe, tasks, duration)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--recipefile', '-r',
                        type=argparse.FileType('r'),
                        required=True)
    parser.add_argument('--envfile', '-e', type=argparse.FileType('r'))
    parser.add_argument('--task', '-t', type=str, nargs='+', default=[])
    parser.add_argument('--loop', '-l', type=bool, default=False)
    parser.add_argument('--debug', type=bool, default=False)
    parser.add_argument('--show', type=bool, default=False)
    parser.add_argument('--duration', type=float, default=0.0)
    d = parser.parse_args()
    _set_logger(d.debug)
    if d.envfile is not None:
        envs = yaml.load(d.envfile)
    else:
        envs = {}
    if not d.loop:
        job(envs, d.recipefile, d.task, d.show, d.duration)
    else:
        job(envs, d.recipefile, d.task, d.show, d.duration, loop=True)


if __name__ == '__main__':
    main()



from collections import OrderedDict
from contextlib import closing
import io

import jinja2
from yaml.dumper import Dumper
from yaml.loader import SafeLoader
from yaml.resolver import BaseResolver


def parse_ordered_yaml(infile):
    def ordering(loader, node):
        return OrderedDict(loader.construct_pairs(node))
    loader = SafeLoader(infile)
    try:
        tag = BaseResolver.DEFAULT_MAPPING_TAG
        loader.add_constructor(tag, ordering)
        return loader.get_single_data()
    finally:
        loader.dispose()


def ordered_dump(doc):
    s = io.StringIO()

    def ordering(dumper, doc):
        tag = BaseResolver.DEFAULT_MAPPING_TAG
        return dumper.represent_mapping(tag, doc.items())

    dumper = Dumper(s, default_flow_style=False)
    dumper.add_representer(OrderedDict, ordering)
    dumper.open()
    with closing(dumper):
        dumper.represent(doc)
    return s.getvalue()


def parse_recipe(infile, config):
    source = infile.read()
    template = jinja2.Template(source)
    rendered_source = template.render(**config)
    recipe = parse_ordered_yaml(rendered_source)
    return recipe



import json
import subprocess
import unittest


class TestRecipeTask(unittest.TestCase):
    def setUp(self):
        httpbin = subprocess.Popen([
            'python', '-m', 'httpbin.core',
            '--host', '127.0.0.1',
            '--port', '4000'],
            stderr=subprocess.PIPE)
        httpbin.stderr.readline()
        self.httpbin = httpbin

    def tearDown(self):
        httpbin = self.httpbin
        httpbin.stderr.close()
        httpbin.kill()
        httpbin.wait()

    def run_recipe(self, taskname):
        c = subprocess.run([
            'python', '-m',
            'bamboo_crawler',
            '-r', 'tests/httpbin/recipe.yml',
            '-t', taskname],
            stdout=subprocess.PIPE)
        self.assertEqual(c.returncode, 0)
        return c

    def test_constant_inputter_test(self):
        c = self.run_recipe('constant_inputter_test')
        expect = b'abc1234'
        self.assertEqual(c.stdout, expect)

    def test_constant_inputter_with_metadata(self):
        c = self.run_recipe('constant_inputter_with_metadata')
        expect = b'abc1234'
        self.assertEqual(c.stdout, expect)

    def test_fetch_task(self):
        c = self.run_recipe('fetch_task')
        expect = b'User-agent: *\nDisallow: /deny\n'
        self.assertEqual(c.stdout, expect)

    def test_user_agent(self):
        c = self.run_recipe('user_agent')
        j = json.loads(c.stdout)
        expect = 'Testing User Agent'
        self.assertEqual(j['user-agent'], expect)



from collections import defaultdict
from datetime import date
from datetime import datetime
from datetime import time
import re
from urllib.parse import parse_qs
from urllib.parse import urlparse

import lxml.html
import requests
from types import MappingProxyType

from .. import interface


class HTTPCrawler(interface.Crawler):
    def __init__(self, headers=MappingProxyType({})):
        self.headers = headers

    def crawl(self, url, *, context=None):
        resp = requests.get(url, headers=self.headers)
        yield resp.content


class XPathScraper(interface.Scraper):
    def __init__(self, xpaths):
        self.xpaths = xpaths

    def scrape(self, data, *, context=None):
        elements = lxml.html.fromstring(data)
        j = {key: elements.xpath(xpath)
             for key, xpath in self.xpaths.items()}
        yield j


class SingleXPathScraper(interface.Scraper):
    def __init__(self, xpath):
        self.xpath = xpath

    def scrape(self, data, *, context=None):
        element = lxml.html.fromstring(data)
        elements = element.xpath(self.xpath)
        k = [str(x) for x in elements]
        yield from k


class CSSSelectorScraper(interface.Scraper):
    def __init__(self, selectors):
        self.selectors = selectors

    def scrape(self, data, *, context=None):
        elements = lxml.html.fromstring(data)
        j = {key: self.__select(elements, p_selector)
             for key, p_selector in self.selectors.items()}
        yield j

    @classmethod
    def __select(cls, elements, p_selector):
        if isinstance(p_selector, (str, bytes)):
            return [e.text_content() for e in elements.cssselect(p_selector)]
        selector, attribute = p_selector
        return [e.attrib[attribute] for e in elements.cssselect(selector)]


class MixedHTMLScraper(interface.Scraper):
    def __init__(self, *,
                 targets=MappingProxyType({})):
        self.targets = targets
        xpaths = {key: target['xpath']
                  for key, target in targets.items()
                  if 'xpath' in target}
        css = {key: target['css']
               for key, target in targets.items()
               if 'css' in target}
        self.xpath_scraper = XPathScraper(xpaths=xpaths)
        self.cssselector_scraper = CSSSelectorScraper(selectors=css)

    def scrape(self, data, *, context=None):
        d1 = list(self.xpath_scraper.scrape(data))[0]
        d2 = list(self.cssselector_scraper.scrape(data))[0]
        yield self.__merge(d1, d2)

    @classmethod
    def __merge(cls, *scraped_data_list):
        d = defaultdict(list)
        for data in scraped_data_list:
            for key, values in data.items():
                d[key].extend(values)
        return dict(d)


class NullProcessor(interface.Processor):
    def process(self, data, *, context=None):
        yield data


class PythonProcessor(interface.Processor):
    def __init__(self, *, mappers):
        self.mappers = mappers

    @classmethod
    def __safe_eval(self, code, data, metadata):

        def extract_digit(data):
            return ''.join(x for x in data if x.isdigit())

        allowed_functions = {
            'int': int,
            'float': float,
            'str': str,
            'extract_digit': extract_digit,
            'urlparse': urlparse,
            'parse_qs': parse_qs,
            'date': date,
            'time': time,
            'datetime': datetime,
            'max': max,
            'all': all,
            'any': any,
            'divmod': divmod,
            'sorted': sorted,
            'ord': ord,
            'chr': chr,
            'bin': bin,
            'sum': sum,
            'pow': pow,
            'len': len,
            'range': range,
            'map': map,
            're': re,
        }
        globals_ = {'__builtins__': allowed_functions}
        locals_ = {'_': data, 'meta': metadata}
        try:
            return eval(code, globals_, locals_)
        except Exception:
            return None

    def process(self, data, *, context=None):
        yield {key: self.__safe_eval(code, data, metadata=context.metadata)
               for key, code in self.mappers.items()}


__all__ = ['HTTPCrawler', 'XPathScraper', 'SingleXPathScraper',
           'CSSSelectorScraper', 'NullProcessor']



import json
import subprocess
import unittest


class TestScraping(unittest.TestCase):
    def run_recipe(self, taskname):
        c = subprocess.run([
            'python', '-m',
            'bamboo_crawler',
            '-r', 'tests/scraping/recipe.yml',
            '-t', taskname],
            stdout=subprocess.PIPE)
        self.assertEqual(c.returncode, 0)
        return c

    def test_mixed_scraper(self):
        c = self.run_recipe('mixed_scrape')
        j = json.loads(c.stdout)
        self.assertEqual(j['x'][0], 'test_message01')
        self.assertEqual(j['y'][0], 'test_message02')

    def test_python_processor(self):
        c = self.run_recipe('python_processor')
        j = json.loads(c.stdout)
        self.assertEqual(j['x'], 'abc')
        self.assertEqual(j['y'], 'xxxyyyzzz')



#!/usr/bin/env python

import subprocess

import pip
from setuptools import find_packages
import setuptools.command.build_py as build_py
from setuptools import setup


def _read_requirements(filename):
    reqs_obj = pip.req.parse_requirements(filename,
                                          session=pip.download.PipSession())
    reqs_str = [str(ir.req) for ir in reqs_obj]
    return reqs_str


class BuildReadMeCommand(build_py.build_py):

    def run(self):
        subprocess.run([
            'pandoc',
            '--from', 'markdown',
            '--to', 'rst',
            '--output', 'README.rst',
            'README.md'
        ])


class PersistentPipSyncCommand(build_py.build_py):

    def run(self):
        for filename in ('requirements.txt', 'dev-requirements.txt'):
            subprocess.run(['pip', 'install', '-r', filename])
        for filename in ('requirements.in', 'dev-requirements.in'):
            subprocess.run(['pip-compile', filename])
        subprocess.run([
            'pip-sync',
            'dev-requirements.txt',
            'requirements.txt',
        ])


# https://docs.python.org/3.6/distutils/setupscript.html#additional-meta-data
setup(
    name='bamboo-crawler',
    version='0.1.0',
    description='Hobby Crawler (yet)',
    long_description='''\
bamboo-crawler
============

bamboo-crawler enables continuous growth in crawling task like a bamboo forest.

.. note:: This package is still roughly developing stage.
   So backward compatibility is very fragile at current.
    ''',
    author='Yui Kitsu',
    author_email='kitsuyui+github@kitsuyui.com',
    url='https://github.com/kitsuyui/bamboo-crawler',
    packages=find_packages(),
    install_requires=_read_requirements('requirements.txt'),
    extras_require={
        'dev': _read_requirements('dev-requirements.txt'),
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
    ],
    test_suite='tests',
    tests_require=['nose'],
    scripts=['bamboo_crawler/__main__.py'],
    entry_points={'console_scripts': [
        'bamboo = bamboo_crawler.cli:main',
    ]},
    cmdclass={
        'dev_build_readme': BuildReadMeCommand,
        'dev_persistent_pip_sync': PersistentPipSyncCommand,
    },
    license=open('./LICENSE', 'r').read(),
)
